#include <fftw.h>
#include <stdio.h>
#include <assert.h>
#include <math.h>

/* destroys obuf... */
static void dozoom (fftw_complex *obuf, fftw_complex *nbuf,
		    int oxsize, int oysize,
		    int nxsize, int nysize,
		    fftwnd_plan oplan, fftwnd_plan nplan)
{
  printf("transforming input from time domain to frequency domain..."); fflush (stdout);
  fftwnd_one (oplan, obuf, NULL);
  printf("done\n");

  { /* preparing for IDFT here. note that imaginary parts are negated here !!! */
    int i;
    double mult, negmult;

    mult = ((double)1) / ((double)(oxsize * oysize));
    negmult = -mult;
    for (i=0; i<oxsize*oysize; ++i) {
      obuf [i].re *= mult;
      obuf [i].im *= negmult;
    }
  }

  { /* create target spectrum from original spectrum */

    /* U group (left upper) */
    int urxsize, urysize; /* real */

    /* K group (lower right) */
    int krxsize, krysize; /* real */

    int x, y;

    { /* calculate sizes of 4 windows. */
      /* U group (left upper) */
      int unxsize, unysize; /* new */
      int uoxsize, uoysize; /* old */
      
      /* K group (lower right) */
      int knxsize, knysize; /* new */
      int koxsize, koysize; /* old */
      unxsize = nxsize / 2 + 1;
      unysize = nysize / 2 + 1;
      
      uoxsize = oxsize / 2 + 1;
      uoysize = oysize / 2 + 1;
      
      urxsize = unxsize < uoxsize ? unxsize : uoxsize;
      urysize = unysize < uoysize ? unysize : uoysize;
      
      knxsize = nxsize - unxsize;
      knysize = nysize - unysize;
      
      koxsize = oxsize - uoxsize;
      koysize = oysize - uoysize;
      
      krxsize = knxsize < koxsize ? knxsize : koxsize;
      krysize = knysize < koysize ? knysize : koysize;
    }

    for (y=0; y<nysize; ++y) {
      for (x=0; x<nxsize; ++x) {
	if ( (x<urxsize) && (y<urysize) ) {
	  nbuf [y*nxsize + x] = obuf[y*oxsize + x];
	} else if ( (x>=nxsize-krxsize) && (y<urysize) ) {
	  nbuf [y*nxsize + x] = obuf[y*oxsize + (x-nxsize+oxsize)];
	} else if ( (x<urxsize) && (y>=nysize-krysize) ) {
 	  nbuf [y*nxsize + x] = obuf[(y-nysize+oysize)*oxsize + x];
	} else if ( (x>=nxsize-krxsize) && (y>=nysize-krysize) ) {
 	  nbuf [y*nxsize + x] = obuf[(y-nysize+oysize)*oxsize + (x-nxsize+oxsize)];
	} else {
	  nbuf [y*nxsize + x].re = 0;
	  nbuf [y*nxsize + x].im = 0;
	}
      }
    }
  }

  printf("transforming output from frequency domain to time domain..."); fflush (stdout);
  fftwnd_one (nplan, nbuf, NULL);
  printf("done\n");
}

int main(int argc, char *argv [])
{
  int oxsize, oysize;
  int nxsize, nysize;

  FILE *x;
  int i, component;
  unsigned char *picbuf[3];
  unsigned char *imm;
  fftw_complex *opic, *npic;
  fftwnd_plan oplan, nplan;
  double zoom;

  printf("Usage: superzoom input.rgb inputxsize inputysize output.rgb scale\n");
  
  /* get size of old image */
  oxsize=atoi(argv[2]);
  oysize=atoi(argv[3]);

  /* get size of new image */
  zoom=atof(argv[5]);
  nxsize = (int) (oxsize * zoom + 0.5);
  nysize = (int) (oysize * zoom + 0.5);
  printf("nxsize=%d nysize=%d\n", nxsize, nysize);

  { /* calculate max size and get picbuf */
    int maxxsize, maxysize;
    maxxsize = oxsize > nxsize ? oxsize : nxsize;
    maxysize = oysize > nysize ? oysize : nysize;
    for (component=0; component<3; ++component)
      assert ((picbuf[component] = (unsigned char *) malloc(maxxsize * maxysize)) != NULL);
  }

  /* load sourcefile into the three picbufs */
  assert ((imm = (unsigned char *) malloc(oxsize * oysize * 3)) != NULL);
  assert ((x = fopen (argv[1], "rb")) != NULL);
  assert (fread (imm, oxsize * oysize * 3, 1, x) == 1);
  assert (fclose(x) == 0);
  for (i=0; i < oxsize * oysize; ++i)
    for (component=0; component<3; ++component)
      picbuf[component][i]= imm[i*3+component];
  free (imm);
  
  /* create fftw plans */
  printf("creating oplan..."); fflush (stdout);
  oplan = fftw2d_create_plan(oysize, oxsize, FFTW_FORWARD, FFTW_ESTIMATE | FFTW_IN_PLACE);
  printf("done\n");
  printf("creating nplan..."); fflush (stdout);
  nplan = fftw2d_create_plan(nysize, nxsize, FFTW_FORWARD, FFTW_ESTIMATE | FFTW_IN_PLACE);
  printf("done\n");

  /* allocate opic and npic fftw buffers */
  assert ((opic = (fftw_complex *) malloc(sizeof(fftw_complex) * oxsize * oysize)) != NULL);
  assert ((npic = (fftw_complex *) malloc(sizeof(fftw_complex) * nxsize * nysize)) != NULL);

  /* for each color channel in picbuf */
  for (component=0; component<3; ++component) {
    /* copy it into the opic fftw buffer */
    int i, v;
    for (i=0; i<oxsize*oysize; ++i) {
      opic[i].re = ((double)picbuf[component][i]) / ((double)255);
      opic[i].im = 0;
    }
    /* zoom it into the npic fftw buffer */
    dozoom (opic, npic, oxsize, oysize, nxsize, nysize, oplan, nplan);
    /* copy result back into the picbuf */
    for (i=0; i<nxsize*nysize; ++i) {
      v = npic[i].re * ((double)255) + ((double)0.5);
      if (v>255) v=255;
      if (v<0) v=0;
      picbuf[component][i] = v;
    }
  }

  /* free fftw buffers and fftw plans */
  free (npic);
  free (opic);
  fftwnd_destroy_plan (nplan);
  fftwnd_destroy_plan (oplan);
  
  /* write output to file */
  assert ((imm = (unsigned char *) malloc(nxsize * nysize * 3)) != NULL);
  for (i=0; i < nxsize * nysize; ++i)
    for (component = 0; component < 3; ++component)
      imm[i*3+component] = picbuf[component][i];
  assert ((x = fopen (argv[4], "wb")) != NULL);
  assert (fwrite (imm, nxsize * nysize * 3, 1, x) == 1);
  assert (fclose(x) == 0);
  free (imm);

  /* free picbuffers */
  for (component=0; component<3; ++component) free(picbuf[component]);

  return 0;
}
